/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "entity.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "world/world.h"
#include "physics/physics.h"

entity_t ENTITIES[ENTITY_COUNT_MAX] = {0};

entitycallback_t ENTITY_CALLBACKS[ENTITY_TYPE_COUNT] = {
  {NULL}, // ENTITY_TYPE_NULL
  {playerNPCInit, playerNPCUpdate}, // ENTITY_TYPE_PLAYER
  {npcInit, npcUpdate} // ENTITY_TYPE_NPC
};

void entityInit(entity_t *entity, const entitytype_t type) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertTrue(type != ENTITY_TYPE_NULL, "Entity type NULL");
  assertTrue(type < ENTITY_TYPE_COUNT, "Entity type out of bounds");
  assertNotNull(
    ENTITY_CALLBACKS[type].init,
    "Entity type has no init callback"
  );
  
  memoryZero(entity, sizeof(entity_t));

  entity->type = type;
  ENTITY_CALLBACKS[type].init(entity);
}

void entityUpdate(entity_t *entity) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertTrue(entity->type != ENTITY_TYPE_NULL, "Entity type NULL");
  assertTrue(entity->type < ENTITY_TYPE_COUNT, "Entity type out of bounds");
  assertNotNull(
    ENTITY_CALLBACKS[entity->type].update,
    "Entity type has no update callback"
  );

  ENTITY_CALLBACKS[entity->type].update(entity);

  if(
    entity->vx == FIXED248(0, 0) &&
    entity->vy == FIXED248(0, 0)
  ) return;

  fixed248_t newX = entity->x + entity->vx;
  fixed248_t newY = entity->y + entity->vy;

  // Check for collisions with the world boundaries

  // Check for collisions with tiles

  // Check for collisions with other entities
  fixed248_t halfTileW = FIXED248(TILE_WIDTH_HEIGHT / 2, 0);
  fixed248_t halfTileH = FIXED248(TILE_WIDTH_HEIGHT / 2, 0);
  fixed248_t selfCircR = halfTileW;
  fixed248_t selfCircX = newX - halfTileW;
  fixed248_t selfCircY = newY - halfTileH;

  for(entity_t *otherEntity = ENTITIES; otherEntity < ENTITIES + ENTITY_COUNT_MAX; ++otherEntity) {
    // Skip self and null entities
    if(otherEntity == entity || otherEntity->type == ENTITY_TYPE_NULL) continue;

    fixed248_t otherCircX = otherEntity->x - halfTileW;
    fixed248_t otherCircY = otherEntity->y - halfTileH;
    fixed248_t otherCircR = halfTileW;

    collisionresult_t collision = physicsCheckCircleCircle(
      selfCircX, selfCircY, selfCircR,
      otherCircX, otherCircY, otherCircR
    );
    if(!collision.hit) continue;

    // Collision with entity detected. Slide out of collision.
    fixed248_t slideX = fx248Mulf32(collision.normalX, fx248Tof32(collision.depth));
    fixed248_t slideY = fx248Mulf32(collision.normalY, fx248Tof32(collision.depth));
    newX -= slideX;
    newY -= slideY;

    // Update selfCircX/Y for next collision check
    selfCircX = newX - halfTileW;
    selfCircY = newY - halfTileH;
  }

  entity->x = newX;
  entity->y = newY;
}